/* cache.hh: Generic cache with LRU removing implementation.
 *
 * Copyright (c) 2007,2008 Jonne Lehtinen <jms@modeemi.fi>
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the Do What The F*ck You Want To
 *  Public License, Version 2, as published by Sam Hocevar. See
 *  COPYING file for more details.
 */

#ifndef ECLI_CACHE_HH
#define ECLI_CACHE_HH

#include <list>
#include <map>

namespace ecli
{

template< typename key_type, typename data_type >
class cache
{

public:
    cache( unsigned int max_size = 10 )
        : MAX_SIZE( max_size ), data_(), order_(), last_( data_.end() ) {}
    ~cache() {}

    void
    add( key_type const& key, data_type const& data );

    data_type
    get( key_type const& key );

    bool
    has( key_type const& key );

private:

    typedef std::list< key_type > access_order;
    typedef typename access_order::iterator order_iter;
    typedef std::pair< data_type, order_iter > data_t;
    typedef std::map< key_type, data_t > data_map;

    unsigned int const MAX_SIZE;
    data_map data_;
    access_order order_;
    typename data_map::iterator last_;
};

template< typename key_type, typename data_type >
void
cache< key_type, data_type >::add( key_type const& key, data_type const& data )
{
    std::pair< typename data_map::iterator, bool > ins( 
        data_.insert( typename data_map::value_type(key, data_t(data,order_.end())) ) );
    if( ins.second )
    {
        if( data_.size() > MAX_SIZE )
        {
            typename data_map::iterator i = data_.find( order_.front() );
            data_.erase( i );
            order_.pop_front();
        }
        order_.push_back( key );
        ins.first->second.second = --order_.end();
    }
}

template< typename key_type, typename data_type >
data_type
cache< key_type, data_type >::get( key_type const& key )
{
    if( has( key ) ) {
        order_.splice( order_.end(), order_, last_->second.second );
        last_->second.second = --order_.end();
        return last_->second.first;
    }
    return data_type();
}

template< typename key_type, typename data_type >
bool
cache< key_type, data_type >::has( key_type const& key )
{
    if( last_ != data_.end() && last_->first == key ) {
        return true;
    }
    typename data_map::iterator i = data_.find( key );
    if( i != data_.end() )
    {
        last_ = i;
        return true;
    }
    return false;
}

}

#endif // ECLI_CACHE_HH
