/* client.cc: XMMS2 client implementation.
 *
 * Copyright (c) 2006 Jonne Lehtinen <jms@modeemi.fi>
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the Do What The F*ck You Want To
 *  Public License, Version 2, as published by Sam Hocevar. See
 *  COPYING file for more details.
 */

#include <xmmsclient/xmmsclient++-glib.h>

#include <cstdlib>
#include <iostream>

#include <glibmm/ustring.h>
#include <gtkmm/main.h>
#include <gdkmm/pixbufloader.h>
#include <gdkmm/pixbuf.h>

#include "client.hh"
#include "window_holder.hh"
#include "config.hh"
#include "skin.hh"

gx2osd::client::client()
    : client_( "gx2osd" ),
      win_( get_window_holder() ), curid_( 0 ), cover_cache_() {

    client_.connect( std::getenv("XMMS_PATH") );
    win_.set_status_no_show( client_.playback.getStatus() );

    client_.setMainloop( new Xmms::GMainloop( client_.getConnection() ) );
    client_.playback.currentID()
        ( Xmms::bind( &client::song_changed, this ) );

    client_.playback.broadcastCurrentID()
        ( Xmms::bind( &client::song_changed, this ) );
    client_.medialib.broadcastEntryChanged()
        ( Xmms::bind( &client::entry_changed, this ) );
    client_.playback.broadcastStatus()
        ( Xmms::bind( &window_holder::set_status, &win_ ) );
    client_.setDisconnectCallback( boost::bind( &client::dc_handler, this ) );

}

gx2osd::client::~client() {
}

bool
gx2osd::client::entry_changed( const unsigned int& id ) {
    if( curid_ == id ) {
        // no it didn't really change but this would have the same effect..
        song_changed( id );
    }
    return true;
}

bool
gx2osd::client::song_changed( const unsigned int& id ) {
    curid_ = id;
    client_.medialib.getInfo( id )( Xmms::bind( &client::info_get, this ) );
    return true;
}

bool 
gx2osd::client::error_handler( const std::string& error,
                               const Xmms::PropDict& info ) {

    std::cerr << "Error: " << error << std::endl;
    win_.set_text_and_show( info );
    return false;

}

bool
gx2osd::client::info_get( const Xmms::PropDict& info ) {

    if( info.contains( "picture_front" ) ||
        info.contains( "picture_front_thumbnail" ) ) {

        std::string hash;

        if( info.contains( "picture_front_thumbnail" ) ) {
            hash = info.get< std::string >( "picture_front_thumbnail" );
        }
        else {
            hash = info.get< std::string >( "picture_front" );
        }

        Glib::RefPtr< Gdk::Pixbuf > image( cover_cache_.get( hash ) );
        if( image ) {
            win_.set_cover( image, info );
        }
        else {

            client_.bindata.retrieve( hash )
                ( boost::bind( &client::cover_get, this, _1, hash, info ),
                  boost::bind( &client::error_handler, this, _1, info ) );
        }

    }
    else {

        win_.set_text_and_show( info );

    }

    return true;
}

bool
gx2osd::client::cover_get( const Xmms::bin& cover,
                           const std::string& cover_hash,
                           const Xmms::PropDict& info ) {

    Glib::RefPtr< Gdk::Pixbuf > image;
    try {

        Glib::RefPtr<Gdk::PixbufLoader> loader = Gdk::PixbufLoader::create();

        loader->write( cover.data(), cover.size() );
        loader->close();

        image = scale_pixbuf( loader->get_pixbuf() );

    }
    catch ( Glib::Error& e ) {
        std::clog << "Could not load image: " << e.what() << std::endl;
        image = get_config().get_skin()->no_cover_image();
    }

    cover_cache_.add( cover_hash, image );
    win_.set_cover( image, info );
    return false;

}

Glib::RefPtr< Gdk::Pixbuf >
gx2osd::client::scale_pixbuf( Glib::RefPtr< Gdk::Pixbuf > const& pixbuf ) const
{
    const int width = pixbuf->get_width();
    const int height = pixbuf->get_height();
    int dest_width = 100;
    int dest_height = 100;
    double ratio = width / static_cast< double >(height);

    if( width > height ) {
        dest_height = static_cast< int >(100 / ratio);
    }
    else if( height > width ) {
        dest_width = static_cast< int >(100 * ratio);
    }

    return pixbuf->scale_simple( dest_width, dest_height, Gdk::INTERP_BILINEAR );
}

void
gx2osd::client::dc_handler() {
    std::clog << "Disconnected..." << std::endl;
    Gtk::Main::quit();
}
