#ifndef PLAYLIST_LANGUAGE_EVALUATOR_HH
#define PLAYLIST_LANGUAGE_EVALUATOR_HH

#include <string>
#include <iostream>
#include <map>
#include <algorithm>
#include <iterator>
#include <xmmsclient/xmmsclient++/dict.h>
#include <boost/spirit/tree/common_fwd.hpp>
#include <boost/typeof/typeof.hpp>
#include <boost/optional.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/bind.hpp>

#include "playlist_language.hh"

namespace playlist_language
{

template< typename T >
class evaluator_impl : public evaluator
{
public: 
    typedef boost::spirit::tree_parse_info< T > parse_info;

    evaluator_impl( parse_info& tree );
    ~evaluator_impl() {}

    void
    set_info( Xmms::Dict const& info )
    {
        info_ = &info;
    }

    void
    run( std::ostream& os = std::cout );

    void
    run( Xmms::Dict const& info, std::ostream& os = std::cout )
    {
        set_info( info );
        run( os );
    }

    bool
    add_function( std::string const& name, func_t func );

private:

    typedef typename boost::spirit::tree_match< T >::const_tree_iterator
        const_tree_iterator;

    parse_info parse_tree_;
    std::map< std::string, variable_type > builtin_id_;
    std::map< std::string, variable_type > variables_;
    std::map< std::string, func_t > functions_;

    bool eval_done_;
    Xmms::Dict const* info_;

    void
    eval_expression( const_tree_iterator it, std::ostream& os );

    bool
    eval_bool( const_tree_iterator it );

    return_t
    eval_func( const_tree_iterator it, std::ostream& os );

    bool
    eval_not( const_tree_iterator it );

    variable_type
    eval_calc( const_tree_iterator it );

    bool
    expr_to_bool( const_tree_iterator it );

    std::string
    expr_to_string( const_tree_iterator it );

    variable_type
    expr_to_val( const_tree_iterator it );

    return_t
    func_pad( arguments const& args, std::ostream& );

    return_t
    func_emit( arguments const& args, std::ostream& os );

    return_t
    func_done( arguments const&, std::ostream& );

    return_t
    func_set( arguments const& args, std::ostream& );

    variable_type
    lookup_id( std::string const& id )
    {
        BOOST_AUTO( it, variables_.find( id ) );
        if( it != variables_.end() ) {
            return it->second;
        }
        else if( (it = builtin_id_.find( id )) != builtin_id_.end() ) {
            return it->second;
        }
        else {
            return std::string("");
        }
    }

};

template< typename T >
evaluator_impl< T >::evaluator_impl( parse_info& tree )
    : evaluator(), parse_tree_( tree ), builtin_id_(), variables_(),
      eval_done_( false ), info_( 0 )
{
    builtin_id_[ "CR" ] = "\n";
    functions_[ "emit" ] = boost::bind( &evaluator_impl< T >::func_emit, this,
                                        _1, _2 );
    functions_[ "done" ] = boost::bind( &evaluator_impl< T >::func_done, this,
                                        _1, _2 );
    functions_[ "set" ] = boost::bind( &evaluator_impl< T >::func_set, this,
                                       _1, _2 );
    functions_[ "pad" ] = boost::bind( &evaluator_impl< T >::func_pad, this,
                                       _1, _2 );
}

template< typename T >
void
evaluator_impl< T >::run( std::ostream& os )
{
    if( info_ ) {
        eval_done_ = false;
        eval_expression( parse_tree_.trees.begin(), os );
    }
}

template< typename T >
bool
evaluator_impl< T >::add_function( std::string const& name, func_t func )
{
    return functions_.insert( std::make_pair(name, func) ).second;
}

template< typename T >
void
evaluator_impl< T >::eval_expression( const_tree_iterator it,
                                      std::ostream& os )
{
    BOOST_AUTO( child_it_begin, it->children.begin() );
    if( it->value.id() == playlist_language::id::CMP ) {
        if( eval_bool( it ) ) {
            child_it_begin += 2;
        }
        else {
            return;
        }
    }
    else if( it->value.id() == playlist_language::id::NOT ) {
        if( eval_not( child_it_begin ) ) {
            child_it_begin += 1;
        }
        else {
            return;
        }
    }
    else if( it->value.id() == playlist_language::id::FUNCTION ) {
        eval_func( it, os );
        return;
    }
    else if( it->value.id() == playlist_language::id::ID ) {
        // TODO: hmm ?
        return;
    }
    else if( it->value.id() == playlist_language::id::TAG ) {
        std::string tag(it->value.begin() + 1, it->value.end());
        if( !info_->contains( tag ) ) {
            return;
        }
    }
    for( const_tree_iterator i = child_it_begin;
         i != it->children.end() && !eval_done_; ++i ) {
        eval_expression( i, os );
    }
}

template< typename T >
bool
evaluator_impl< T >::eval_bool( const_tree_iterator it )
{
    std::string op(it->value.begin(), it->value.end());
    // TODO: check for argument count? is it needed?
    if( op == "&&" ) {
        bool bool1 = expr_to_bool( it->children.begin() );
        bool bool2 = false;

        if( it->children[1].value.id() == playlist_language::id::CMP ) {
            bool2 = eval_bool( it->children.begin() + 1 );
        }
        else {
            bool2 = expr_to_bool( it->children.begin() + 1 );
        }
        return bool1 && bool2;
    }
    else if( op == "==" ) {
        // TODO: this is a quick hack, doesn't work properly
        std::string val1( expr_to_string( it->children.begin() ) );
        std::string val2( expr_to_string( it->children.begin() + 1 ) );
        return val1 == val2;
    }
    else if( op == "<" ) {
        const_tree_iterator child = it->children.begin();
        variable_type val1( expr_to_val( child ) );
        variable_type val2( expr_to_val( child + 1 ) );
        return val1 < val2;
    }
    return false;
}

template< typename T >
bool
evaluator_impl< T >::eval_not( const_tree_iterator it )
{
    std::string op(it->value.begin(), it->value.end());
    if( it->value.id() == playlist_language::id::ID ) {
        return variables_.find( op ) == variables_.end();
    }
    std::clog << "Unimplemented or unsupported type"
                 "for operator not" << std::endl;
    return false;
    // TODO: more types?
}

struct add : public boost::static_visitor< evaluator::variable_type >
{

    evaluator::variable_type
    operator()( std::string const& lhs, std::string const& rhs ) const
    {
        return lhs + rhs;
    }

    template< typename T >
    evaluator::variable_type
    operator()( std::string const& lhs, T const& rhs ) const
    {
        return lhs + boost::lexical_cast< std::string >( rhs );
    }

    template< typename T >
    evaluator::variable_type
    operator()( T const& lhs, std::string const& rhs ) const
    {
        return boost::lexical_cast< std::string >( lhs ) + rhs;
    }

    template< typename T, typename U >
    evaluator::variable_type
    operator()( T const& lhs, U const& rhs ) const
    {
        return lhs + rhs;
    }

    template< typename T >
    evaluator::variable_type
    operator()( T const& lhs, T const& rhs ) const
    {
        return lhs + rhs;
    }
};

struct subtract : public boost::static_visitor< evaluator::variable_type >
{

    evaluator::variable_type
    operator()( std::string const&, std::string const& ) const
    {
        // TODO: undefined?
        return std::string();
    }

    template< typename T >
    evaluator::variable_type
    operator()( std::string const&, T const& ) const
    {
        return std::string();
    }

    template< typename T >
    evaluator::variable_type
    operator()( T const&, std::string const& ) const
    {
        return std::string();
    }

    template< typename T, typename U >
    evaluator::variable_type
    operator()( T const& lhs, U const& rhs ) const
    {
        return lhs - rhs;
    }

    template< typename T >
    evaluator::variable_type
    operator()( T const& lhs, T const& rhs ) const
    {
        return lhs - rhs;
    }
};

struct divide : public boost::static_visitor< evaluator::variable_type >
{

    evaluator::variable_type
    operator()( std::string const&, std::string const& ) const
    {
        // TODO: undefined?
        return std::string();
    }

    template< typename T >
    evaluator::variable_type
    operator()( std::string const&, T const& ) const
    {
        return std::string();
    }

    template< typename T >
    evaluator::variable_type
    operator()( T const&, std::string const& ) const
    {
        return std::string();
    }

    template< typename T, typename U >
    evaluator::variable_type
    operator()( T const& lhs, U const& rhs ) const
    {
        if( rhs != 0 ) {
            return lhs / rhs;
        }
        return std::string();
    }

    template< typename T >
    evaluator::variable_type
    operator()( T const& lhs, T const& rhs ) const
    {
        if( rhs != 0 ) {
            return lhs / rhs;
        }
        return std::string();
    }
};

struct multiply : public boost::static_visitor< evaluator::variable_type >
{

    evaluator::variable_type
    operator()( std::string const&, std::string const& ) const
    {
        // TODO: undefined?
        return std::string();
    }

    template< typename T >
    evaluator::variable_type
    operator()( std::string const&, T const& ) const
    {
        return std::string();
    }

    template< typename T >
    evaluator::variable_type
    operator()( T const&, std::string const& ) const
    {
        return std::string();
    }

    template< typename T, typename U >
    evaluator::variable_type
    operator()( T const& lhs, U const& rhs ) const
    {
        return lhs * rhs;
    }

    template< typename T >
    evaluator::variable_type
    operator()( T const& lhs, T const& rhs ) const
    {
        return lhs * rhs;
    }
};

template< typename T >
evaluator::variable_type
evaluator_impl< T >::expr_to_val( const_tree_iterator it )
{
    evaluator::variable_type val;

    using namespace playlist_language::id;
    std::string temp( it->value.begin(), it->value.end() );
    if( it->value.id() == CALC ) {
        val = eval_calc( it );
    }
    if( it->value.id() == STRING ) {
        val = temp;
    }
    else if( it->value.id() == INT ) {
        val = boost::lexical_cast< int >( temp );
    }
    else if( it->value.id() == DOUBLE ) {
        val = boost::lexical_cast< double >( temp );
    }
    else if( it->value.id() == ID ) {
        val = lookup_id( temp );
    }
    else if( it->value.id() == TAG ) {
        temp = temp.substr(1);
        if( info_->contains( temp ) ) {
            Xmms::Dict::Variant var( (*info_)[ temp ] );
            if( var.type() != typeid( std::string ) ) {
                val = boost::get< int >( var );
            }
            else {
                val = boost::get< std::string >( var );
            }
        }
    }
    return val;
}

template< typename T >
evaluator::variable_type
evaluator_impl< T >::eval_calc( const_tree_iterator it )
{
    evaluator::variable_type val1( expr_to_val( it->children.begin() ) );
    evaluator::variable_type val2( expr_to_val( it->children.begin() + 1 ) );
    switch( *it->value.begin() ) {
        case '+': {
            return boost::apply_visitor( add(), val1, val2 );
        } break;
        case '*': {
            return boost::apply_visitor( multiply(), val1, val2 );
        } break;
        case '-': {
            return boost::apply_visitor( subtract(), val1, val2 );
        } break;
        case '/': {
            return boost::apply_visitor( divide(), val1, val2 );
        } break;
    }
    return 0;
}

template< typename T >
evaluator::return_t
evaluator_impl< T >::eval_func( const_tree_iterator it, std::ostream& os )
{
    typedef typename boost::spirit::tree_match< T >::node_t node;
    std::string func_name( it->value.begin(), it->value.end() );
    BOOST_AUTO( func_it, functions_.find( func_name ) );
    if( func_it != functions_.end() ) {
        arguments args;
        args.reserve( it->children.size() );
        for( const_tree_iterator n = it->children.begin();
             n != it->children.end(); ++n )
        {
            using namespace playlist_language::id;
            if( n->value.id() == ID ) {
                // FIXME: ugly workaround for a function which actually
                // needs the id as the first argument and not the value
                // the id holds.
                std::string temp(n->value.begin(), n->value.end());
                if( func_name != "set" ) {
                    args.push_back( lookup_id( temp ) );
                }
                else {
                    if( n != it->children.begin() ) {
                        args.push_back( lookup_id( temp ) );
                    }
                    else {
                        args.push_back( temp );
                    }
                }
            }
            else if( n->value.id() == FUNCTION ) {
                return_t val = eval_func( n, os );
                if( val ) {
                    args.push_back( *val );
                }
            }
            else {
                args.push_back( expr_to_val( n ) );
            }
        }
        return func_it->second( args, os );
    }
    return return_t();
}

template< typename T >
bool
evaluator_impl< T >::expr_to_bool( const_tree_iterator it )
{
    if( it->value.id() == playlist_language::id::TAG ) 
    {
        std::string foo( it->value.begin() + 1, it->value.end() );
        return info_->contains( foo );
    }
    // TODO: implement the rest
    return false;
}

template< typename T >
std::string
evaluator_impl< T >::expr_to_string( const_tree_iterator it )
{
    return boost::lexical_cast< std::string >( expr_to_val( it ) );
}

template< typename T >
evaluator::return_t
evaluator_impl< T >::func_pad( arguments const& args, std::ostream& )
{
    if( args.size() != 1 ) {
        // FIXME: real error handling?
        std::clog << "pad(): invalid number of arguments\n";
        return return_t();
    }

    unsigned int len = 0;
    using boost::get;
    if( args[0].type() == typeid( int ) ) {
        len = get< int >( args[0] );
    }
    else if( args[0].type() == typeid( std::string ) ) {
        /* len() for utf8 strings...
        for( std::string::const_iterator o = t.begin();
             o != t.end(); ++o ) {
            if( (*o & 0xC0) != 0x80 ) {
                ++len;
            }
        }
        */
        len = get< std::string >( args[0] ).size();
    }
    else {
        // FIXME: handle error
        std::clog << "pad(): unsupported type\n";
    }

    std::string temp( len, ' ' );
    return return_t( temp );
}

template< typename T >
evaluator::return_t
evaluator_impl< T >::func_emit( arguments const& args, std::ostream& os )
{
    std::ostream_iterator< arguments::value_type, char > out( os );
    std::copy( args.begin(), args.end(), out );
    return return_t();
}

template< typename T >
evaluator::return_t
evaluator_impl< T >::func_done( arguments const&, std::ostream& )
{
    eval_done_ = true;
    return return_t();
}

template< typename T >
evaluator::return_t
evaluator_impl< T >::func_set( arguments const& args, std::ostream& )
{
    if( args.size() != 2 ) {
        // FIXME: real error handling
        std::clog << "set(): invalid number of arguments\n";
        return return_t();
    }
    if( args[0].type() != typeid( std::string ) ) {
        // FIXME: real error handling
        std::clog << "set(): expected id as first argument\n";
        return return_t();
    }
    
    variables_[ boost::get< std::string >( args[0] ) ] = args[1];
    return return_t();
}

}

#endif // EVALUATOR_HH
