/* main_window.cc: OSD Window implementation.
 *
 * Copyright (c) 2006 Jonne Lehtinen <jms@modeemi.fi>
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the Do What The F*ck You Want To
 *  Public License, Version 2, as published by Sam Hocevar. See
 *  COPYING file for more details.
 */

#include <gtk/gtk.h>
#include <gtkmm/window.h>
#include <gtkmm/label.h>
#include <gdkmm/screen.h>
#include <gdkmm/pixbuf.h>
#include <gtkmm/image.h>
#include <glibmm/ustring.h>

#include "main_window.hh"
#include "config.hh"
#include "scroller.hh"
#include "window_holder.hh"

#include <iostream>

#include <xmmsclient/xmmsclient++.h>
#include "boost/foreach.hpp"


gx2osd::main_window::main_window( Glib::RefPtr< Gdk::Screen > screen ) 
    : Gtk::Window( Gtk::WINDOW_POPUP ),
      labels_(),
      fixed_pane_(),
      span_end_( "</span>" ),
      cover_(),
      status_(),
      screen_( screen ),
      skin_( get_config().get_skin() ) {

    if( screen_ ) {
        set_screen( screen_ );
    }
        // TODO: find a way to do this in gtkmm... or prove not possible yet.
        gtk_window_set_accept_focus( GTK_WINDOW( gobj() ), FALSE );
    set_decorated( false );
    set_keep_above( true );
    set_events( Gdk::BUTTON_PRESS_MASK );

    load_skin();

    if( !get_config().get_position().empty() ) {

        gx2osd::config::positions_container::size_type pos =
            get_config().get_index_of_position( get_config().get_position() );

        int screen_width = screen_->get_width();
        int screen_height = screen_->get_height();
        Glib::RefPtr<Gdk::Pixbuf> buf( skin_->background() );
        int width = buf->get_width();
        int height = buf->get_height();
        switch( pos ) {
            case 0: {
                move( 0, 0 );
            } break;
            case 1: {
                move( (screen_width / 2 - width / 2), 0 );
            } break;
            case 2: {
                move( screen_width - width, 0 );
            } break;
            case 3: {
                move( 0, (screen_height / 2 - height / 2) );
            } break;
            case 4: {
                set_position( Gtk::WIN_POS_CENTER );
            } break;
            case 5: {
                move( screen_width - width,
                      (screen_height / 2 - height / 2) );
            } break;
            case 6: {
                move( 0, screen_height - height );
            } break;
            case 7: {
                move( (screen_width / 2 - width / 2),
                      screen_height - height );
            } break;
            case 8: {
                move( screen_width - width,
                      screen_height - height );
            } break;
        }

    }
    else {
        move( get_config().get_xpos(), get_config().get_ypos() );
    }

    add( fixed_pane_ );
    gx2osd::get_window_holder().add_window( this );

}

gx2osd::main_window::~main_window() {

}

void
gx2osd::main_window::set_text( const Xmms::PropDict& info,
                               unsigned int extra_width )
{

    std::ostringstream os;
    BOOST_FOREACH( label_tuple temp, labels_ )
    {
        os.str("");
        temp.get<2>().get_evaluator()->run( info, os );
        temp.get<0>()->set_markup( temp.get<1>()
                               + Glib::Markup::escape_text( os.str() )
                               + span_end_ );
        temp.get<0>()->set_size_request( temp.get<2>().get_width() + extra_width );
    }

}

void
gx2osd::main_window::set_text_and_show( const Xmms::PropDict& info ) {

    unsigned int i = 100;
    Glib::RefPtr< Gdk::Pixbuf > temp = skin_->no_cover_image();
    if( temp ) {
        background_.set( skin_->background() );
        cover_.set( temp );
        i = 0;
    }
    else if( (temp = skin_->no_cover_background()) ) {
        cover_.clear();
        background_.set( temp );
    }
    set_text( info, i );

}

void
gx2osd::main_window::set_cover( const Glib::RefPtr< Gdk::Pixbuf >& cover,
                                const Xmms::PropDict& info ) {

    unsigned int i = 100;
    if( skin_->has_cover() ) {
        cover_.set( cover );
        i = 0;
    }
    set_text( info, i );

}

bool
gx2osd::main_window::set_status_no_show( const Xmms::Playback::Status& status ) {
    switch( status ) {
        case Xmms::Playback::STOPPED: {
            status_.set( skin_->stopped_image() );
            return get_config().get_stop();
            break;
        }
        case Xmms::Playback::PLAYING: {
            status_.set( skin_->playing_image() );
            return get_config().get_play();
            break;
        }
        case Xmms::Playback::PAUSED: {
            status_.set( skin_->paused_image() );
            return get_config().get_pause();
            break;
        }
    }
    return true;
}

void
gx2osd::main_window::load_skin() {

    Glib::RefPtr<Gdk::Pixbuf> buf( skin_->background() );

    if( buf ) {

        Glib::RefPtr<Gdk::Bitmap> mask;
        Glib::RefPtr<Gdk::Pixmap> pixmap;

        buf->render_pixmap_and_mask( pixmap, mask, 128 );

        resize( buf->get_width(), buf->get_height() );

        mask.clear();

        // The render_pixmap_and_mask doesn't take a screen object 
        // and bitmaps must be screen specific in X so we must do manually 
        // that part of the function. Pretty much copied from gtk+ code.
        GdkBitmap* m = gdk_pixmap_new( screen_->get_root_window()->gobj(),
                                       buf->get_width(), buf->get_height(), 1 );
        gdk_pixbuf_render_threshold_alpha( buf->gobj(), m,  0, 0, 0, 0,
                                           buf->get_width(), buf->get_height(), 128 );
        gtk_widget_shape_combine_mask( GTK_WIDGET(gobj()), m, 0, 0 );
        g_object_unref( m );

        background_.set( buf );

    }

    // This must be added first, otherwise it will get on top of others
    fixed_pane_.add( background_ );

    gx2osd::skin::label_map const& labels = skin_->get_labels();
    BOOST_FOREACH( gx2osd::skin::label_map::value_type i, labels )
    {
        label_tuple temp( scroller_ptr( new eclidgets::scroller() ),
                          Glib::ustring(), i.second );
        set_label_stuff( *temp.get<0>(), temp.get<2>(), temp.get<1>() );
        labels_.push_back( temp );
        fixed_pane_.put( *temp.get<0>(), temp.get<2>().get_position().x,
                                         temp.get<2>().get_position().y );
    }

    fixed_pane_.put( cover_, skin_->no_cover_image().get_position().x, 
                             skin_->no_cover_image().get_position().y );
    fixed_pane_.put( status_, skin_->get_status_position().x, 
                              skin_->get_status_position().y );
    
    cover_.set( skin_->no_cover_image() );
    status_.set( skin_->paused_image() );
    
}

void
gx2osd::main_window::set_label_stuff( eclidgets::scroller& gtklabel,
                                      const gx2osd::label& label,
                                      Glib::ustring& markup ) {

    Pango::FontDescription font = gtklabel.get_font_description();
    font.set_size( label.get_size() * Pango::SCALE );
    if( !label.get_font().empty() ) {
        font.set_family( label.get_font() );
    }
    gtklabel.set_font_description( font );

    markup = "<span weight=\"";

    if( label.get_bold() ) {
        markup += "bold\"";
    }
    else {
        markup += "normal\"";
    }
    markup += " color=\"" + label.get_color() + "\">";

    gtklabel.set_size_request( label.get_width() );

}

bool
gx2osd::main_window::on_button_press_event( GdkEventButton* event )
{
    Gtk::Window::on_button_press_event( event );
    hide();
    return true;
}
