#include "playlist_language.hh"
#include "evaluator.hh"
#include <boost/spirit/core.hpp>
#include <boost/spirit/utility/lists.hpp>
#include <boost/spirit/iterator/file_iterator.hpp>
#include <boost/spirit/tree/ast.hpp>

using namespace boost::spirit;

namespace 
{

using namespace playlist_language::id;
struct pll_grammar : public grammar< pll_grammar >
{

template< typename ScannerT >
struct definition
{
    definition( pll_grammar const& /* self */ )
    {
        pll = *(block >> *eol_p) >> end_p;
        block = root_node_d[ cmpexpr ] >> no_node_d[ eol_p ] >>
                +statement >> no_node_d[ (*eol_p || end_p) ];

        cmpexpr = notexpr >> *(no_node_d[ *blank_p ] >>
                  root_node_d[ str_p( "&&" ) ||
                               str_p( "==" ) ||
                               ch_p('<') ] >>
                  no_node_d[ *blank_p ] >> cmpexpr);

        notexpr = root_node_d[ch_p('!')] >> no_node_d[ *blank_p ] >>
                  token_node_d[ arg ]
                | arg;

        statement = (no_node_d[+blank_p] >> function >>
                     no_node_d[(eol_p || end_p)]);

        arg = no_node_d[*blank_p] >>
              (function || str || calc || id || tag || number ) >>
              no_node_d[*blank_p];

        tag = leaf_node_d[ ':' >> +alpha_p ];
        id = leaf_node_d[ +alpha_p ];
        str = no_node_d[ch_p('\"')] >>
              leaf_node_d[(*(~cntrl_p - '\"'))] >>
              no_node_d[ch_p('\"')];

        function = root_node_d[ leaf_node_d[+alpha_p] ] >>
                   inner_node_d[ch_p('(') >>
                   infix_node_d[ *(arg >> *(',' >> arg)) ] >> ch_p(')')];

        number = real | integer;
        integer = int_p;
        real = strict_real_p;

        factor = number || id
               | inner_node_d[ch_p('(') >> calc >> ch_p(')')]
               | (root_node_d[ch_p('(')] >> factor);

        term   = factor >>
                 *( (no_node_d[*blank_p] >> root_node_d[ch_p('*')] >>
                     no_node_d[*blank_p] >> factor)
                  | (no_node_d[*blank_p] >> root_node_d[ch_p('/')] >>
                     no_node_d[*blank_p] >> factor)
                 );

        calc   = term >>
                 *( (no_node_d[*blank_p] >> root_node_d[ch_p('+')] >>
                     no_node_d[*blank_p] >> term)
                  | (no_node_d[*blank_p] >> root_node_d[ch_p('-')] >>
                     no_node_d[*blank_p] >> term)
                 );

        BOOST_SPIRIT_DEBUG_NODE( pll );
        BOOST_SPIRIT_DEBUG_NODE( block );
        BOOST_SPIRIT_DEBUG_NODE( statement );
        BOOST_SPIRIT_DEBUG_NODE( arg );
        BOOST_SPIRIT_DEBUG_NODE( id );
        BOOST_SPIRIT_DEBUG_NODE( str );
        BOOST_SPIRIT_DEBUG_NODE( tag );
        BOOST_SPIRIT_DEBUG_NODE( function );
        BOOST_SPIRIT_DEBUG_NODE( notexpr );
        BOOST_SPIRIT_DEBUG_NODE( factor );
        BOOST_SPIRIT_DEBUG_NODE( term );
        BOOST_SPIRIT_DEBUG_NODE( calc );

    }

    rule< ScannerT > pll;
    rule< ScannerT > statement;
    rule< ScannerT > number;

    rule< ScannerT, parser_context<>, parser_tag< BLOCK > > block;
    rule< ScannerT, parser_context<>, parser_tag< FUNCTION > > function;
    rule< ScannerT, parser_context<>, parser_tag< STRING > > str;
    rule< ScannerT, parser_context<>, parser_tag< ID > > id;
    rule< ScannerT, parser_context<>, parser_tag< TAG > > tag;
    rule< ScannerT, parser_context<>, parser_tag< ARG > > arg;
    rule< ScannerT, parser_context<>, parser_tag< NOT > > notexpr;
    rule< ScannerT, parser_context<>, parser_tag< CMP > > cmpexpr;
    rule< ScannerT, parser_context<>, parser_tag< INT > > integer;
    rule< ScannerT, parser_context<>, parser_tag< DOUBLE > > real;
    rule< ScannerT, parser_context<>, parser_tag< CALC > > factor;
    rule< ScannerT, parser_context<>, parser_tag< CALC > > term;
    rule< ScannerT, parser_context<>, parser_tag< CALC > > calc;

    rule< ScannerT > const&
    start() const { return pll; }
};

};

typedef file_iterator< char > file_it;

}

playlist_language::evaluator_ptr
playlist_language::parse_file( std::string const& file )
{
    pll_grammar pll;
    file_it first( file );
    file_it last = first.make_end();
    tree_parse_info< file_it > tree = ast_parse( first, last, pll );
    if( tree.full ) {
        return evaluator_ptr( new evaluator_impl< file_it >( tree ) );
    }
    else {
        return evaluator_ptr();
    }
}

playlist_language::evaluator_ptr
playlist_language::parse_data( std::string const& data )
{
    pll_grammar pll;
    tree_parse_info<> tree = ast_parse( data.c_str(), pll );
    if( tree.full ) {
        return evaluator_ptr( new evaluator_impl< char const* >( tree ) );
    }
    else {
        return evaluator_ptr();
    }
}
