/* scroller.hh: Scrolling text widget header
 *
 * Copyright (c) 2006-2008 Jonne Lehtinen <jms@modeemi.fi>
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the Do What The F*ck You Want To
 *  Public License, Version 2, as published by Sam Hocevar. See
 *  COPYING file for more details.
 */

#ifndef ECLIPSERS_WIDGET_SCROLLER_HH
#define ECLIPSERS_WIDGET_SCROLLER_HH

#include <gtkmm/widget.h>
#include <pangomm/layout.h>
#include <glibmm/ustring.h>

namespace eclidgets
{

class scroller : public Gtk::Widget
{

public:

    enum scrolling
    {
        SCROLL_SMART,  // Scrolls when the text is longer than width of widget.
        SCROLL_ALWAYS, // Scrolls always
        SCROLL_NEVER   // Never scrolls.
    };

    enum scroll_direction
    {
        SCROLL_TO_LEFT = -1,
        SCROLL_TO_RIGHT = 1
    };

    // Constructor with rather sane default values.
    scroller( Glib::ustring const& text = "",
              scrolling scroll = SCROLL_SMART,
              scroll_direction direction = SCROLL_TO_LEFT,
              Gtk::Justification default_justify = Gtk::JUSTIFY_CENTER,
              unsigned int interval = 50, unsigned int multiplier = 1,
              unsigned int wait_before_scroll = 1000,
              int gap = 50 );

    virtual ~scroller();

    // Some getters for different values

    Glib::ustring
    get_text() const;

    scrolling
    get_scroll() const;

    scroll_direction
    get_direction() const;

    Gtk::Justification
    get_justification() const;

    void
    get_speed( unsigned int& interval, unsigned int& multiplier ) const;

    int
    get_gap() const;

    unsigned int
    get_wait_before_scroll() const;

    Pango::FontDescription
    get_font_description() const;

    // Setters
    void
    set_text( Glib::ustring const& str );

    void
    set_markup( Glib::ustring const& markup );

    // Set scrolling type
    void
    set_scroll( scrolling scroll );

    // Set scrolling direction
    void
    set_direction( scroll_direction direction );

    void
    set_justification( Gtk::Justification justify );

    // Set speed of scrolling,
    // interval is the time between movement calls in milliseconds,
    // multiplier is how many pixels the text moves on each movement call.
    void
    set_speed( unsigned int interval, unsigned int multiplier = 1 );

    // Sets the gap between the head and the tail of the scrolling text so
    // that they won't be too close to eachother.
    // Negative value is possible but may cause some funny behaviour.
    void
    set_gap( int gap );

    // Set how many milliseconds to wait before starting scroll.
    // This helps reading the text.
    void
    set_wait_before_scroll( unsigned int wait_before_scroll );

    // See Gtkmm docs on this, a lot of things have something similar.
    void
    set_font_description( Pango::FontDescription const& desc );

    void
    set_size_request( int width = -1, int height = -1 );

protected:

    virtual void
    on_realize();

    virtual bool
    on_expose_event( GdkEventExpose* event );

    virtual void
    on_size_allocate( Gtk::Allocation& allocation );

private:

    Glib::RefPtr< Pango::Layout > layout_;
    Glib::RefPtr< Gdk::GC > gc_;

    int pos_;
    Gtk::Justification justify_;
    scrolling scroll_;
    scroll_direction direction_;

    bool scrolling_;
    int width_;

    unsigned int interval_;
    unsigned int multiplier_;
    unsigned int wait_before_scroll_;
    int gap_;

    sigc::connection conn_;
    sigc::connection wait_;

    bool
    timer();

    void
    determine_scroll();

    bool
    start_scroll();

    void
    text_changed();

};

}

#endif // ECLIPSERS_WIDGET_SCROLLER_HH
