#include <iostream>
#include <list>

#include <boost/typeof/typeof.hpp>
#include <boost/function.hpp>
#include <boost/foreach.hpp>
#include <boost/bind.hpp>

#include "detail/shortcut_impl.hh"
#include "key_combination.hh"

#include <gdk/gdkx.h>
#include <gdk/gdkdisplay.h>
#include <gdk/gdkkeys.h>
#include <gdk/gdkscreen.h>
#include <gdk/gdkdisplaymanager.h>

#include <X11/Xlib.h>

namespace ecli
{

namespace detail
{

struct impl_data
{
    Display* xdisp_;
    std::list< Window > xwin_;

    struct key
    {
        guint code;
        guint modifiers;
        boost::function< void() > callback;
    };
    std::list< key > keys_;
    boost::function< bool( XKeyEvent* ) > cb_;

    bool
    handle_event( XKeyEvent* xkey );
};

}

}

static GdkFilterReturn
filter( GdkXEvent* xevent, GdkEvent* /*event*/, gpointer data )
{
    BOOST_AUTO( e, (XEvent*)xevent );
    if( e->type == KeyPress ) {
        BOOST_AUTO( cb, (boost::function< bool( XKeyEvent* ) >*)data );
        if( (*cb)( &(e->xkey) ) ) {
            return GDK_FILTER_REMOVE;
        }
    }
    return GDK_FILTER_CONTINUE;
}

ecli::detail::shortcut_impl::shortcut_impl()
    : data( new impl_data() )
{
    data->xdisp_ = GDK_DISPLAY();

    GdkDisplay* disp = gdk_display_get_default();
    gint screens = gdk_display_get_n_screens( disp );
    for( gint i = 0; i < screens; ++i )
    {
        GdkScreen* screen = gdk_display_get_screen( disp, i );
        data->xwin_.push_back( GDK_WINDOW_XWINDOW( gdk_screen_get_root_window( screen ) ) );
    }
    data->cb_ = boost::bind( &ecli::detail::impl_data::handle_event, data, _1 );

    gdk_window_add_filter( NULL, &filter, &data->cb_ );
}

ecli::detail::shortcut_impl::~shortcut_impl()
{
    BOOST_FOREACH( impl_data::key i, data->keys_ )
    {
        BOOST_FOREACH( Window win, data->xwin_ )
        {
            XUngrabKey( data->xdisp_, i.code, i.modifiers, win );
        }
    }
    delete data;
}

static bool grabbed;

static int
x11_grab_error_handler(Display*, XErrorEvent*)
{
    grabbed = false;
    return 0;
}

bool
ecli::detail::shortcut_impl::register_shortcut( ecli::key_combination const& key,
                                                boost::function< void() > const& callback )
{
    GdkKeymapKey* keys = 0;
    gint n_keys = 0;
    guint code = 0;
    if( gdk_keymap_get_entries_for_keyval( NULL, key.get_key(), &keys, &n_keys ) )
    {
        if( n_keys > 0 )
        {
            code = keys[0].keycode;
        }
        g_free( keys );
    }
    else {
        return false;
    }
    grabbed = true;
    XErrorHandler old_handler = XSetErrorHandler(x11_grab_error_handler);
    BOOST_FOREACH( Window win, data->xwin_ )
    {
        XGrabKey( data->xdisp_, code, key.get_modifiers(), win,
                  False, GrabModeAsync, GrabModeAsync );
    }
    impl_data::key temp = { code, key.get_modifiers(), callback };
    data->keys_.push_back( temp );
    XSync( data->xdisp_, False );
    XSetErrorHandler( old_handler );
    return grabbed;
}

void
ecli::detail::shortcut_impl::unregister_shortcut( ecli::key_combination const& key )
{

}

bool
ecli::detail::impl_data::handle_event( XKeyEvent* xkey )
{
    BOOST_FOREACH( key i, keys_ )
    {
        if( xkey->keycode == i.code &&
            xkey->state == i.modifiers ) {
            i.callback();
            return true;
        }
    }
    return false;
}
