/* skin.cc: skin format implementation.
 *
 * Copyright (c) 2006 Jonne Lehtinen <jms@modeemi.fi>
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the Do What The F*ck You Want To
 *  Public License, Version 2, as published by Sam Hocevar. See
 *  COPYING file for more details.
 */

#include "skin.hh"
#include "config.hh"
#include "image.hh"
#include "label.hh"
#include "playlist_language.hh"

#include "config.h"

#include <gdkmm/pixbuf.h>

#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLString.hpp>

#include <xercesc/dom/DOMDocument.hpp>
#include <xercesc/dom/DOMNode.hpp>
#include <xercesc/dom/DOMNodeList.hpp>
#include <xercesc/dom/DOMElement.hpp>

#include <xercesc/sax/SAXParseException.hpp>
#include <xercesc/sax/ErrorHandler.hpp>

#include "boost/filesystem/path.hpp"
#include "boost/filesystem/operations.hpp"
#include "boost/algorithm/string/trim.hpp"

#include <string>
#include <sstream>
#include <iostream>

std::ostream& operator<<( std::ostream& stream, const XMLCh* const c ) {
    char* temp = xercesc::XMLString::transcode( c );
    stream << temp;
    xercesc::XMLString::release( &temp );
    return stream;
}

class err_handler : public xercesc::ErrorHandler {

    public:
        err_handler() {}
        virtual ~err_handler() {}

        void warning( const xercesc::SAXParseException& e ) {
            std::cerr << "Warning (" << e.getLineNumber() << ":"
                      << e.getColumnNumber() << "): "
                      << e.getMessage() << std::endl;
        }

        void error( const xercesc::SAXParseException& e ) {
            std::ostringstream ss;
            ss << "Error (" << e.getLineNumber() << ":"
               << e.getColumnNumber() << "): "
               << e.getMessage();
            throw std::runtime_error( ss.str() );
        }

        void fatalError( const xercesc::SAXParseException& e ) {
            std::ostringstream ss;
            ss << "Fatal error (" << e.getLineNumber() << ":"
               << e.getColumnNumber() << "): "
               << e.getMessage();
            throw std::runtime_error( ss.str() );
        }

        void resetErrors() {}

};

class XMLChDeleter {

    public:
        XMLChDeleter() {}
        ~XMLChDeleter() {}

        void operator()( XMLCh*& p ) {
            xercesc::XMLString::release( &p );
        }

};

gx2osd::skin::skin( const std::string& name )
    : noncopyable(),
      name_( name ),
      skin_path_( get_config().get_config_dir() / "skins" ),
      has_cover_( false ),
      background_(),
      no_cover_background_(),
      no_cover_image_(),
      paused_image_(), 
      stopped_image_(), 
      playing_image_(),
      labels_(),
      status_position_() {

    if( !fs::exists( (skin_path_ / name) / (name + ".xml") ) ) {
#ifdef SHAREDIR
        skin_path_ = SHAREDIR "/skins";
        if( !fs::exists( (skin_path_ / name) / (name + ".xml") ) ) {
#endif
        throw std::invalid_argument( "Could not find skin." );
#ifdef SHAREDIR
        }
#endif
    }

    xercesc::XMLPlatformUtils::Initialize();

    boost::shared_ptr< xercesc::XercesDOMParser > parser_;

    err_handler err;
    parser_.reset( new xercesc::XercesDOMParser() );
    parser_->setErrorHandler( &err );
    parser_->setValidationScheme( xercesc::XercesDOMParser::Val_Always );
    parser_->setValidationSchemaFullChecking( true );
    parser_->setDoNamespaces( true );
    parser_->setDoSchema( true );
#ifdef SHAREDIR
    parser_->setExternalNoNamespaceSchemaLocation( 
            (fs::path(SHAREDIR) / "skins/skin.xsd").string().c_str() );
#endif

    parser_->parse( ((skin_path_ / name) / (name + ".xml")).string().c_str() );

    xercesc::DOMDocument* doc = parser_->getDocument();

    using xercesc::XMLString;

    xmlch background_tag( XMLString::transcode("background"), XMLChDeleter() );
    xmlch cover_tag( XMLString::transcode("cover"), XMLChDeleter() );
    xmlch paused_tag( XMLString::transcode("paused"), XMLChDeleter() );
    xmlch stopped_tag( XMLString::transcode("stopped"), XMLChDeleter() );
    xmlch playing_tag( XMLString::transcode("playing"), XMLChDeleter() );

    xmlch image_attr( XMLString::transcode("image"), XMLChDeleter() );
    xmlch nocover_attr( XMLString::transcode("nocover"), XMLChDeleter() );

    load_image( background_, background_tag, doc, image_attr );
    load_image( no_cover_background_, background_tag, doc, nocover_attr);
    load_image( no_cover_image_, cover_tag, doc, image_attr );
    load_image( paused_image_, paused_tag, doc, image_attr );
    load_image( stopped_image_, stopped_tag, doc, image_attr );
    load_image( playing_image_, playing_tag, doc, image_attr );

    load_labels( doc );

    xmlch status_tag( XMLString::transcode("status"), XMLChDeleter() );

    using xercesc::DOMElement;

    DOMElement* el = 0;
    if( (el = find_element( status_tag, doc ).first) ) {
        parse_position( el, status_position_.x, status_position_.y );
    }
    has_cover_ = (find_element( cover_tag, doc ).first != 0);

    parser_.reset();
    xercesc::XMLPlatformUtils::Terminate();

}

gx2osd::skin::~skin() {

}

bool
gx2osd::skin::has_cover() const {
    return has_cover_;
}

gx2osd::image&
gx2osd::skin::background() {
    return background_;
}

gx2osd::image&
gx2osd::skin::no_cover_background()
{
    return no_cover_background_;
}

gx2osd::image&
gx2osd::skin::no_cover_image() {
    return no_cover_image_;
}

gx2osd::image&
gx2osd::skin::paused_image() {
    return paused_image_;
}

gx2osd::image&
gx2osd::skin::stopped_image() {
    return stopped_image_;
}

gx2osd::image&
gx2osd::skin::playing_image() {
    return playing_image_;
}

const gx2osd::position&
gx2osd::skin::get_status_position() const {
    return status_position_;
}

const gx2osd::skin::label_map&
gx2osd::skin::get_labels() const {
    return labels_;
}

void
gx2osd::skin::load_image( gx2osd::image& i, xmlch tag_name,
                          xercesc::DOMDocument*& doc, xmlch attribute ) {

    using xercesc::XMLString;
    using xercesc::DOMElement;

    DOMElement* el = 0;
    if( (el = find_element( tag_name, doc ).first) ) {

        if( el->hasAttribute( &*attribute ) ) {
            std::stringstream ss;
            ss << el->getAttribute( &*attribute );

            std::clog << "\tLoading image: " << ss.str() << std::endl;
            fs::path image_file( skin_path_ / name_ / ss.str() );
            if( !exists( image_file ) ) {
                throw std::invalid_argument( std::string( "Could not find image: " )
                                             + image_file.string() );
            }

            i = Gdk::Pixbuf::create_from_file( image_file.string() );
        }

        unsigned int x = 0;
        unsigned int y = 0;
        parse_position( el, x, y );

        i.set_position( x, y );

    }

}

void gx2osd::skin::load_labels( xercesc::DOMDocument*& doc ) {

    using xercesc::XMLString;
    using xercesc::DOMElement;
    using xercesc::DOMNodeList;

    xmlch tag_name( XMLString::transcode("label"), XMLChDeleter() );

    DOMNodeList* el = 0;
    if( (el = find_element( tag_name, doc ).second) ) {
        for( unsigned int i = 0; i < el->getLength(); ++i ) {
            handle_label( dynamic_cast< DOMElement* >( el->item( i ) ) );
        }
    }

}

void gx2osd::skin::handle_label( xercesc::DOMElement* element )
{
    using xercesc::XMLString;
    using xercesc::DOMElement;
    using xercesc::DOMNodeList;

    if( !element ) {
        return;
    }

    xmlch name_attr( XMLString::transcode("name"), XMLChDeleter() );
    std::stringstream ss;
    ss << element->getAttribute( &*name_attr );

    std::pair< std::map< std::string, label >::iterator, bool > it =
    labels_.insert( std::make_pair( ss.str(), gx2osd::label() ) );
    if( !it.second ) {
        return;
    }
    ss.str("");
    label& l = it.first->second;

    unsigned int x = 0, y = 0;
    parse_position( element, x, y );
    l.set_position( x, y );

    xmlch width_attr( XMLString::transcode("width"), XMLChDeleter() );

    ss << element->getAttribute( &*width_attr );

    unsigned int width = 0;
    ss >> width;

    l.set_width( width );

    xmlch font_elem( XMLString::transcode("font"), XMLChDeleter() );
    xercesc::DOMNodeList* nodes = element->getElementsByTagName( &*font_elem );
    if( nodes && nodes->getLength() == 1 ) {

        DOMElement* el = dynamic_cast< DOMElement* >( nodes->item( 0 ) );

        ss.str("");
        ss.clear();
        ss << el->getTextContent();
        l.set_font( boost::algorithm::trim_copy(ss.str()) );

        std::cout << "\tLoading font attributes... " << std::flush;

        xmlch size_attr( XMLString::transcode("size"), XMLChDeleter() );
        xmlch bold_attr( XMLString::transcode("bold"), XMLChDeleter() );
        xmlch color_attr( XMLString::transcode("color"), XMLChDeleter() );

        if( el->hasAttribute( &*size_attr ) ) {
            std::cout << "size = " << std::flush;
            ss.str("");
            ss.clear();
            ss << el->getAttribute( &*size_attr );
            unsigned int size = 0;
            ss >> size;
            l.set_size( size );
            std::cout << size << "... " << std::flush;
        }
        if( el->hasAttribute( &*bold_attr ) ) {
            std::cout << "bold = " << std::flush;
            ss.str("");
            ss.clear();
            ss << el->getAttribute( &*bold_attr );
            l.set_bold( ss.str() == "true" );
            std::cout << ss.str() << "... " << std::flush;
        }
        if( el->hasAttribute( &*color_attr ) ) {
            std::cout << "color = " << std::flush;
            ss.str("");
            ss.clear();
            ss << el->getAttribute( &*color_attr );
            l.set_color( ss.str() );
            std::cout << ss.str() << "...";
        }
        std::cout << std::endl;

    }

    xmlch content_elem( XMLString::transcode("content"), XMLChDeleter() );
    nodes = element->getElementsByTagName( &*content_elem );
    if( nodes && nodes->getLength() == 1 ) {

        DOMElement* el = dynamic_cast< DOMElement* >( nodes->item( 0 ) );
        std::cout << "\tParsing content definition... " << std::flush;

        ss.str("");
        ss << el->getTextContent();
        std::string line;
        std::string pll;
        std::string::size_type indent = std::string::npos;
        while( std::getline( ss, line ) ) {
            if( indent == std::string::npos ) {
                indent = line.find_first_not_of( " " );
            }
            if( indent != std::string::npos && line.size() >= indent) {
                pll += line.substr( indent ) + "\n";
            }
        }

        playlist_language::evaluator_ptr eval =
            playlist_language::parse_data( pll );
        if( eval ) {
            std::cout << "success!";
            l.set_evaluator( eval );
        }
        else {
            std::cout << "failed!";
            throw std::runtime_error("Failed to parse playlist language");
        }
        std::cout << std::endl;

    }

}

void
gx2osd::skin::parse_position( xercesc::DOMElement*& element,
                              unsigned int& x, unsigned int& y ) {

    using xercesc::XMLString;

    xmlch posit_attr( XMLString::transcode("position"), XMLChDeleter() );

    if( element->hasAttribute( &*posit_attr ) ) {

        std::stringstream ss;
        ss << element->getAttribute( &*posit_attr );

        char temp = 'x';
        ss >> x >> temp >> y;

    }

}

std::pair< xercesc::DOMElement*, xercesc::DOMNodeList* >
gx2osd::skin::find_element( xmlch tag_name, xercesc::DOMDocument*& doc ) {

    using xercesc::DOMNodeList;
    using xercesc::DOMNode;
    using xercesc::DOMElement;

    DOMElement* el = 0;

    DOMNodeList* nodes = doc->getElementsByTagName( &*tag_name );
    if( nodes && nodes->getLength() > 0 ) {

        DOMNode* node = nodes->item( 0 );
        if( node->getNodeType() == DOMNode::ELEMENT_NODE ) {

            el = dynamic_cast< DOMElement* >( node );

        }

    }

    return std::make_pair( el, nodes );

}
