/* window_holder.cc: Proxy class to control multiple windows at once
 *
 * Copyright (c) 2006 Jonne Lehtinen <jms@modeemi.fi>
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the Do What The F*ck You Want To
 *  Public License, Version 2, as published by Sam Hocevar. See
 *  COPYING file for more details.
 */

#include "main_window.hh"
#include "window_holder.hh"
#include "config.hh"
#include <boost/scoped_ptr.hpp>
#include <boost/foreach.hpp>
#include <glibmm/ustring.h>

boost::scoped_ptr< gx2osd::window_holder > gx2osd::window_holder::instance_;

gx2osd::window_holder::~window_holder() {
}

gx2osd::window_holder&
gx2osd::window_holder::get_instance() {
    if( !instance_ ) {
        instance_.reset( new window_holder() );
    }
    return *instance_;
}

void
gx2osd::window_holder::add_window( gx2osd::main_window* window )
{
    windows_.push_back( window );
}

void
gx2osd::window_holder::show_all()
{
    BOOST_FOREACH( gx2osd::main_window* win, windows_ ) {
        win->show_all();
    }
    if( conn_ ) {
        conn_.disconnect();
    }
    conn_ = Glib::signal_timeout().connect( sigc::mem_fun(*this,
                                            &window_holder::hider),
                                            get_config().get_delay() );
}

bool
gx2osd::window_holder::set_status_no_show( const Xmms::Playback::Status& s ) {
    BOOST_FOREACH( gx2osd::main_window* win, windows_ ) {
        win->set_status_no_show( s );
    }
    return true;
}

bool
gx2osd::window_holder::set_status( const Xmms::Playback::Status& s ) {
    bool show = false;
    BOOST_FOREACH( gx2osd::main_window* win, windows_ ) {
        show = win->set_status_no_show( s );
    }
    if( show ) {
        show_all();
    }
    return true;
}

void
gx2osd::window_holder::set_text_and_show( const Xmms::PropDict& info ) {
    BOOST_FOREACH( gx2osd::main_window* win, windows_ ) {
        win->set_text_and_show( info );
    }
    if( get_config().get_song_change() ) {
        show_all();
    }
}

void
gx2osd::window_holder::set_cover( const Glib::RefPtr< Gdk::Pixbuf >& cover,
                                  const Xmms::PropDict& info ) {
    BOOST_FOREACH( gx2osd::main_window* win, windows_ ) {
        win->set_cover( cover, info );
    }
    if( get_config().get_song_change() ) {
        show_all();
    }
}

gx2osd::window_holder::window_holder()
    : boost::noncopyable(),
      windows_() {
}

bool
gx2osd::window_holder::hider()
{
    BOOST_FOREACH( gx2osd::main_window* win, windows_ )
    {
        win->hide();
    }
    return false;
}

gx2osd::window_holder&
gx2osd::get_window_holder() {
    return gx2osd::window_holder::get_instance();
}
